;; The following functions use the Sparse-vector and Sparse-matrix{-operator}
;; machinery in sparse-linear-algebra, 

;; This function defines the Operator that projects an element of the
;; dual of a Linear-finite-element-vector from fine-fes to a
;; Linear-finite-element-vector of coarse-fes.
;; This function uses  the fact that
;; (Vertex-fine vertex) of the coarse triangulation contains
;; the underlying vertex of the fine triangulation, and (Edge-fine e)
;; of the coarse triangulation contains the fine vertex at the
;; midpoint of e in the coarse triangulation.

;; This should be a generic function on any Finite-element-space.

(define (coarse<-fine coarse-space fine-space)
  (with-access
   coarse-space (Finite-element-space triangulation)
   (let ((coarse-triangulation triangulation))
     (let ((pseudo-diagonal-matrix
	    (->Sparse-matrix-operator
	     (instantiate Sparse-matrix                                    ; has Sparse-vector rows
			  rows: (map1                                       ; for each vertex
				 (lambda (v)                               ; make a Sparse-vector
				   (with-access
				    v (Vertex fine)
				    (list (make-Entry (Vertex-index fine) ; underlying vertex has coefficient 1.0
						      1.0))))
				 (Triangulation-vertices->list             ; elements have same indices as vertices
				  (Triangulation-vertices coarse-triangulation)))
			  domain-dimension: (Finite-element-space-neighbors-length fine-space)
			  range-dimension: (Finite-element-space-neighbors-length coarse-space))))
	   (off-diagonal-matrix
	    (->Sparse-matrix-operator                                      ; has indices and coefficients
	     (instantiate Sparse-matrix                                    ; has Sparse-vector rows
			  rows: (map1                                       ; for each vertex
				 (lambda (v)                               ; make a Sparse-vector
				   (with-access
				    v (Vertex fine edges)
				    (sort!                                 ; sort by index
				     (map1 (lambda (e)                ; other vertices have coefficient 0.5
					     (make-Entry (Vertex-index (Edge-fine e))
							 0.5))
					   (Vertex-edges->list edges))
				     (lambda (x y)
				       (FIX (< (Entry-index x) (Entry-index y)))))))
				 (Triangulation-vertices->list             ; elements have same indices as vertices
				  (Triangulation-vertices coarse-triangulation)))
			  domain-dimension: (Finite-element-space-neighbors-length fine-space)
			  range-dimension: (Finite-element-space-neighbors-length coarse-space)))))
       (let ((matrix (Vector-add pseudo-diagonal-matrix off-diagonal-matrix)))
	 (with-access
	  off-diagonal-matrix (Sparse-matrix-operator indices offsets range-dimension)
	  (let ((pseudo-diagonal-indices (Sparse-matrix-operator-indices pseudo-diagonal-matrix))
		(indices indices)
		(offsets offsets)
		(range-dimension range-dimension))
	    (Sparse-matrix-operator-mapping-set!
	     matrix
	     (lambda (v)
	       (declare (fixnum))
	       (let ((pseudo-diagonal-indices pseudo-diagonal-indices)
		     (indices    indices)  ; off-diagonal indices
		     (offsets    offsets)  ; off-diagonal-offsets
		     (data v))
		 (let* ((n range-dimension)
			(result (make-f64vector n)))
		   (let outer ((i (- n 1)) (next-offset (vector-ref offsets n)))
		     (if (< i 0)
			 result
			 (let ((offset (vector-ref offsets i)))
			   (cond ((= (- next-offset offset) 6) ; interior vertices in a refined triangulation
				  (f64vector-set! 
				   result 
				   i 
				   (FLOAT
				    (* 0.5 
				       (+ (+ (+ (f64vector-ref data (indices-ref indices offset))
						(f64vector-ref data (indices-ref indices (FIX (+ offset 1)))))
					     (+ (f64vector-ref data (indices-ref indices (FIX (+ offset 2))))
						(f64vector-ref data (indices-ref indices (FIX (+ offset 3))))))
					  (+ (+ (f64vector-ref data (indices-ref indices (FIX (+ offset 4))))
						(f64vector-ref data (indices-ref indices (FIX (+ offset 5)))))
					     (* 2.0 (f64vector-ref data (indices-ref pseudo-diagonal-indices i))))))))
				  (outer (- i 1) offset))
				 ((= (- next-offset offset) 4) ; boundary vertices in a refined triangulation
				  (f64vector-set! 
				   result 
				   i 
				   (FLOAT
				    (+ (f64vector-ref data (indices-ref pseudo-diagonal-indices i))
				       (* 0.5 
					  (+ (+ (f64vector-ref data (indices-ref indices offset))
						(f64vector-ref data (indices-ref indices (FIX (+ offset 1)))))
					     (+ (f64vector-ref data (indices-ref indices (FIX (+ offset 2))))
						(f64vector-ref data (indices-ref indices (FIX (+ offset 3))))))))))
				  (outer (- i 1) offset))
				 (else
				  (do ((j (- next-offset 1) (- j 1))
				       (inner-result 0.0 (FLOAT
							  (+ inner-result 
							     (f64vector-ref data (indices-ref indices j))))))
				      ((< j offset) 
				       (f64vector-set! result i (FLOAT (+ (f64vector-ref data (indices-ref pseudo-diagonal-indices i))
									  (* 0.5 inner-result))))
				       (outer (- i 1) offset))))))))))))
	    (wrap-matrix-implementation-in-linear-finite-element-operator matrix fine-space coarse-space))))))))

(define (projector->injector m)
  (declare (fixnum))
  (with-access
   m (Linear-finite-element-operator domain-space range-space matrix-implementation)
   (let ((matrix (Operator-adjoint matrix-implementation)))
     (with-access
      matrix (Sparse-matrix-operator indices offsets)
      (let ((indices indices)
	    (offsets offsets))
	(Sparse-matrix-operator-mapping-set! 
	 matrix
	 (lambda (data)
	   (let ((indices indices)
		 (offsets offsets))
	     (let* ((n (vector-length offsets))
		    (result (make-f64vector (- n 1))))
	       (let loop ((i (- n 2)) 
			  (previous-offset (vector-ref offsets (- n 1))))
		 (if (< i 0)
		     result
		     (let ((offset (vector-ref offsets i)))
		       (cond ((= (- previous-offset offset) 2)
			      (f64vector-set! result i (FLOAT (* 0.5 (+ (f64vector-ref data (indices-ref indices offset))
									(f64vector-ref data (indices-ref indices (FIX (+ offset 1))))))))
			      (loop (- i 1) offset))
			     (else
			      (f64vector-set! result i (f64vector-ref data (indices-ref indices offset)))
			      (loop (- i 1) offset))))))))))))
     (wrap-matrix-implementation-in-linear-finite-element-operator matrix range-space domain-space))))
